import logging

from requests_oauthlib import OAuth2Session

from django.shortcuts import redirect, get_object_or_404, render
from django.urls import reverse
from django.http.response import HttpResponseBadRequest

from esi.models import CallbackRedirect, Token
from esi import app_settings
from .decorators import tokens_required


logger = logging.getLogger(__name__)


def sso_redirect(request, scopes=None, return_to=None):
    """
    Generates a :model:`esi.CallbackRedirect` for the specified request.
    Redirects to EVE for login.
    Accepts a view or URL name as a redirect after SSO.
    """
    logger.debug(
        "Initiating redirect of %s session %s",
        request.user, 
        request.session.session_key[:5] if request.session.session_key else '[no key]'
    )
    if scopes is None:
        scopes = list()
    elif isinstance(scopes, str):
        scopes = list([scopes])

    # ensure only one callback redirect model per session
    if request.session.session_key:
        CallbackRedirect.objects\
            .filter(session_key=request.session.session_key).delete()

    # ensure we have a session
    if not request.session.exists(request.session.session_key):
        logger.debug("Creating new session before redirect.")
        request.session.create()

    if return_to:
        url = reverse(return_to)
    else:
        url = request.get_full_path()

    oauth = OAuth2Session(
        app_settings.ESI_SSO_CLIENT_ID, 
        redirect_uri=app_settings.ESI_SSO_CALLBACK_URL, 
        scope=scopes
    )
    redirect_url, state = oauth.authorization_url(app_settings.ESI_OAUTH_LOGIN_URL)

    CallbackRedirect.objects.create(
        session_key=request.session.session_key, state=state, url=url
    )
    logger.debug(
        "Redirecting %s session %s to SSO. Callback will be redirected to %s",
        request.user, 
        request.session.session_key[:5], 
        url
    )
    return redirect(redirect_url)


def receive_callback(request):
    """
    Parses SSO callback, validates, retrieves :model:`esi.Token`, 
    and internally redirects to the target url.
    """
    logger.debug(
        "Received callback for %s session %s", 
        request.user, 
        request.session.session_key[:5]
    )
    # make sure request has required parameters
    code = request.GET.get('code', None)
    state = request.GET.get('state', None)
    try:
        assert code
        assert state
    except AssertionError:
        logger.debug("Missing parameters for code exchange.")
        return HttpResponseBadRequest()

    callback = get_object_or_404(
        CallbackRedirect, state=state, session_key=request.session.session_key
    )
    token = Token.objects.create_from_request(request)
    callback.token = token
    callback.save()
    logger.debug(
        "Processed callback for %s session %s. Redirecting to %s",
        request.user, 
        request.session.session_key[:5], 
        callback.url
    )
    return redirect(callback.url)


def select_token(request, scopes='', new=False):
    """
    Presents the user with a selection of applicable tokens for the requested view.
    """

    @tokens_required(scopes=scopes, new=new)
    def _token_list(r, tokens):
        context = {
            'tokens': tokens,
        }
        return render(r, 'esi/select_token.html', context=context)

    return _token_list(request)



from django.conf import settings

# These are required for SSO to function. 
# Can be left blank if settings.DEBUG is set to True
ESI_SSO_CLIENT_ID = getattr(settings, 'ESI_SSO_CLIENT_ID', None)
ESI_SSO_CLIENT_SECRET = getattr(settings, 'ESI_SSO_CLIENT_SECRET', None)
ESI_SSO_CALLBACK_URL = getattr(settings, 'ESI_SSO_CALLBACK_URL', None)

# Change these to switch to Singularity
ESI_API_DATASOURCE = getattr(settings, 'ESI_API_DATASOURCE', 'tranquility')
ESI_OAUTH_URL = getattr(
    settings, 'ESI_SSO_BASE_URL', 'https://login.eveonline.com/oauth'
)

# Change this to access different revisions of the ESI API by default
ESI_API_VERSION = getattr(settings, 'ESI_API_VERSION', 'latest')

# Enable to force new token creation every callback
ESI_ALWAYS_CREATE_TOKEN = getattr(settings, 'ESI_ALWAYS_CREATE_TOKEN', False)

# Disable to stop caching endpoint responses
ESI_CACHE_RESPONSE = getattr(settings, 'ESI_CACHE_RESPONSE', True)

# Enable/disable verbose info logging
ESI_INFO_LOGGING_ENABLED = getattr(settings, 'ESI_INFO_LOGGING_ENABLED', False)

# Set log level for libraries like bravado and urlib3
ESI_LOG_LEVEL_LIBRARIES = str(getattr(settings, 'ESI_LOG_LEVEL_LIBRARIES', 'INFO'))

# Max size of the connection pool. Increase this setting if you hav more parallel
# threads connected to ESI at the same time.
ESI_CONNECTION_POOL_MAXSIZE = getattr(settings, 'ESI_CONNECTION_POOL_MAXSIZE', 10)

# Max retries on failed connections
ESI_CONNECTION_ERROR_MAX_RETRIES = getattr(
    settings, 'ESI_CONNECTION_ERROR_MAX_RETRIES', 3
)

# Max retries on server errors
ESI_SERVER_ERROR_MAX_RETRIES = getattr(settings, 'ESI_SERVER_ERROR_MAX_RETRIES', 3)

# Backoff factor for retries on server error
ESI_SERVER_ERROR_BACKOFF_FACTOR = getattr(
    settings, 'ESI_SERVER_ERROR_BACKOFF_FACTOR', 0.2
)

# Default timeouts for all requests to ESI. 
# Can be overwritten by passing "timeout" with result()
ESI_REQUESTS_CONNECT_TIMEOUT = getattr(settings, 'ESI_REQUESTS_CONNECT_TIMEOUT', 5)
ESI_REQUESTS_READ_TIMEOUT = getattr(settings, 'ESI_REQUESTS_READ_TIMEOUT', 30)

# These probably won't ever change. Override if needed.
ESI_API_URL = getattr(settings, 'ESI_API_URL', 'https://esi.evetech.net/')
ESI_OAUTH_LOGIN_URL = getattr(
    settings, 'ESI_SSO_LOGIN_URL', ESI_OAUTH_URL + "/authorize/"
)
ESI_TOKEN_URL = getattr(settings, 'ESI_CODE_EXCHANGE_URL', ESI_OAUTH_URL + "/token")
ESI_TOKEN_VERIFY_URL = getattr(
    settings, 'ESI_TOKEN_EXCHANGE_URL', ESI_OAUTH_URL + "/verify"
)
ESI_TOKEN_VALID_DURATION = int(getattr(settings, 'ESI_TOKEN_VALID_DURATION', 1170))
ESI_SPEC_CACHE_DURATION = int(getattr(settings, 'ESI_SPEC_CACHE_DURATION', 3600))

# list of all official language codes supported by ESI
ESI_LANGUAGES = getattr(settings, 'ESI_LANGUAGES', [
    'de', 'en-us', 'fr', 'ja', 'ru', 'zh', 'ko'
])



from datetime import datetime
from hashlib import md5
import json
import logging
from time import sleep
from urllib import parse as urlparse

from bravado.client import SwaggerClient
from bravado import requests_client
from bravado.exception import (
    HTTPBadGateway, HTTPGatewayTimeout, HTTPServiceUnavailable
)
from bravado.swagger_model import Loader
from bravado.http_future import HttpFuture
from bravado_core.spec import Spec, CONFIG_DEFAULTS
from requests.adapters import HTTPAdapter

from django.core.cache import cache

from .errors import TokenExpiredError
from . import app_settings


logger = logging.getLogger(__name__)

_LIBRARIES_LOG_LEVEL = logging.getLevelName(app_settings.ESI_LOG_LEVEL_LIBRARIES)
logging.getLogger('swagger_spec_validator').setLevel(_LIBRARIES_LOG_LEVEL)
logging.getLogger('bravado_core').setLevel(_LIBRARIES_LOG_LEVEL)
logging.getLogger('urllib3').setLevel(_LIBRARIES_LOG_LEVEL)
logging.getLogger('bravado').setLevel(_LIBRARIES_LOG_LEVEL)

SPEC_CONFIG = {'use_models': False}
RETRY_SLEEP_SECS = 1


class CachingHttpFuture(HttpFuture):
    """
    Used to add caching to certain HTTP requests according to "Expires" header
    """    
    def _cache_key(self):
        """
        Generated the key name used to cache responses        
        :return: formatted cache name
        """
        request = self.future.request
        str_hash = md5(
            (
                request.method 
                + request.url 
                + str(request.params) 
                + str(request.data) 
                + str(request.json)
            ).encode('utf-8')
        ).hexdigest()
        return 'esi_%s' % str_hash

    @staticmethod
    def _time_to_expiry(expires):
        """
        Determines the seconds until a HTTP header "Expires" timestamp
        :param expires: HTTP response "Expires" header
        :return: seconds until "Expires" time
        """
        try:
            expires_dt = datetime.strptime(str(expires), '%a, %d %b %Y %H:%M:%S %Z')
            delta = expires_dt - datetime.utcnow()
            return delta.total_seconds()
        except ValueError:
            return 0

    def results(self, **kwargs):
        """Executes the request and returns the response from ESI for the current 
        route. Response will include all pages if there are more available.

        Optional parameters:
        - timeout: timeout for request to ESI in seconds, overwrites default
        """
        results = list()
        headers = None
        # preserve original value
        _also_return_response = self.request_config.also_return_response
        # override to always get the raw response for expiry header
        self.request_config.also_return_response = True  
        
        if "page" in self.operation.params:
            current_page = 1
            total_pages = 1
           
            # loop all pages and add data to output array
            while current_page <= total_pages:  
                self.future.request.params["page"] = current_page                
                # will use cache if applicable
                result, headers = self.result(**kwargs)
                total_pages = int(headers.headers['X-Pages'])  
                # append to results list to be seamless to the client
                results += result  
                current_page += 1
        else:  # it doesn't so just return
            results, headers = self.result(**kwargs)

        # restore original value
        self.request_config.also_return_response = _also_return_response  

        # obey the output 
        if self.request_config.also_return_response:  
            return results, headers
        else:
            return results

    def results_localized(self, languages: list = None, **kwargs) -> dict:
        """Executes the request and returns the response from ESI for all default
        languages and pages (if any). This method returns a dict of all responses 
        with the language code as keys.

        Optional parameters:
        - languages: list of languages to return instead of all default languages
        - timeout: timeout for request to ESI in seconds, overwrites default
        """
        if not languages:
            my_languages = list(app_settings.ESI_LANGUAGES)
        else:
            my_languages = []
            for lang in dict.fromkeys(languages):
                if lang not in app_settings.ESI_LANGUAGES:
                    raise ValueError('Invalid language code: %s' % lang)
                my_languages.append(lang)
        
        return {
            language: self.results(language=language, **kwargs) 
            for language in my_languages
        }

    def result(self, **kwargs):
        """Executes the request and returns the response from ESI. Response will
        include the requested / first page only if there are more pages available.

        Optional parameters:
        - timeout: timeout for request to ESI in seconds, overwrites default
        """
        if 'language' in kwargs.keys():
            # this parameter is not supported by bravado, so we can't pass it on
            self.future.request.params['language'] = str(kwargs.pop('language'))
            
        if 'timeout' not in kwargs:
            kwargs['timeout'] = (
                app_settings.ESI_REQUESTS_CONNECT_TIMEOUT,
                app_settings.ESI_REQUESTS_READ_TIMEOUT
            )
        
        if (
            app_settings.ESI_CACHE_RESPONSE 
            and self.future.request.method == 'GET' 
            and self.operation is not None
        ):           
            cache_key = self._cache_key()
            cached = cache.get(cache_key)
            if cached:
                result, response = cached
                expiry = self._time_to_expiry(str(response.headers['Expires']))
                if expiry < 0:
                    logger.warning(
                        "cache expired by %d seconds, Forcing expiry", expiry
                    )
                    cached = False

            if not cached:
                # preserve original value
                _also_return_response = self.request_config.also_return_response  
                # override to always get the raw response for expiry header
                self.request_config.also_return_response = True
                
                retries = 1
                max_retries = app_settings.ESI_SERVER_ERROR_MAX_RETRIES
                while retries <= max_retries:
                    try:
                        if app_settings.ESI_INFO_LOGGING_ENABLED:
                            params = self.future.request.params
                            logger.info(
                                'Fetching from ESI: %s%s%s',
                                self.future.request.url,
                                f' - language {params["language"]}' 
                                if 'language' in params else '',
                                f' - page {params["page"]}' 
                                if 'page' in params else ''
                            )
                        logger.debug(
                            'ESI request: %s - %s',
                            self.future.request.url,
                            self.future.request.params
                        )
                        logger.debug(
                            'ESI request headers: %s', self.future.request.headers
                        )
                        result, response = super().result(**kwargs)
                        logger.debug(
                            'ESI response status code: %s', response.status_code
                        )
                        logger.debug('ESI response headers: %s', response.headers)
                        logger.debug('ESI response content: %s', response.text)
                        break
                    except (
                        HTTPBadGateway, HTTPGatewayTimeout, HTTPServiceUnavailable
                    ) as ex:
                        if retries < max_retries:
                            logger.warning(
                                "ESI error (Retry: %d/%d)", 
                                retries, 
                                max_retries,
                                exc_info=True
                            )
                            if retries > 1:
                                wait_secs = (
                                    app_settings.ESI_SERVER_ERROR_BACKOFF_FACTOR 
                                    * (2 ** (retries - 1))
                                )
                                sleep(wait_secs)
                            retries += 1
                        else:
                            raise ex
                
                # restore original value
                self.request_config.also_return_response = _also_return_response  

                if 'Expires' in response.headers:
                    expires = self._time_to_expiry(response.headers['Expires'])
                    if expires > 0:
                        cache.set(cache_key, (result, response), expires)

            if self.request_config.also_return_response:
                return result, response
            else:
                return result
        else:
            return super().result(**kwargs)


requests_client.HttpFuture = CachingHttpFuture


class TokenAuthenticator(requests_client.Authenticator):
    """
    Adds the authorization header containing access token, if specified.
    Sets ESI datasource to tranquility or singularity.
    """

    def __init__(self, token=None, datasource=None):
        host = urlparse.urlsplit(app_settings.ESI_API_URL).hostname
        super().__init__(host)
        self.token = token
        self.datasource = datasource

    def apply(self, request):
        if self.token and self.token.expired:
            if self.token.can_refresh:
                self.token.refresh()
            else:
                raise TokenExpiredError()
        request.headers['Authorization'] = \
            'Bearer ' + self.token.access_token if self.token else None
        request.params['datasource'] = \
            self.datasource or app_settings.ESI_API_DATASOURCE
        return request


def build_cache_name(name):
    """
    Cache key name formatter
    :param name: Name of the spec dict to cache, usually version
    :return: String name for cache key
    :rtype: str
    """
    return 'esi_swaggerspec_%s' % name


def cache_spec(name, spec):
    """
    Cache the spec dict
    :param name: Version name
    :param spec: Spec dict
    :return: True if cached
    """
    return cache.set(
        build_cache_name(name), spec, app_settings.ESI_SPEC_CACHE_DURATION
    )


def build_spec_url(spec_version):
    """
    Generates the URL to swagger.json for the ESI version
    :param spec_version: Name of the swagger spec version, like latest or v4
    :return: URL to swagger.json for the requested spec version
    """
    return urlparse.urljoin(app_settings.ESI_API_URL, spec_version + '/swagger.json')


def get_spec(name, http_client=None, config=None):
    """
    :param name: Name of the revision of spec, eg latest or v4
    :param http_client: Requests client used for retrieving specs
    :param config: Spec configuration - see Spec.CONFIG_DEFAULTS
    :return: :class:`bravado_core.spec.Spec`
    """
    http_client = http_client or requests_client.RequestsClient()

    def load_spec():
        loader = Loader(http_client)
        return loader.load_spec(build_spec_url(name))

    spec_dict = cache.get_or_set(
        build_cache_name(name), load_spec, app_settings.ESI_SPEC_CACHE_DURATION
    )
    config = dict(CONFIG_DEFAULTS, **(config or {}))
    return Spec.from_dict(spec_dict, build_spec_url(name), http_client, config)


def build_spec(base_version, http_client=None, **kwargs):
    """
    Generates the Spec used to initialize a SwaggerClient, 
    supporting mixed resource versions
    :param http_client: :class:`bravado.requests_client.RequestsClient`
    :param base_version: Version to base the spec on. 
    Any resource without an explicit version will be this.
    :param kwargs: Explicit resource versions, by name (eg Character='v4')
    :return: :class:`bravado_core.spec.Spec`
    """
    base_spec = get_spec(base_version, http_client=http_client, config=SPEC_CONFIG)
    if kwargs:
        for resource, resource_version in kwargs.items():
            versioned_spec = get_spec(
                resource_version, http_client=http_client, config=SPEC_CONFIG
            )
            try:
                spec_resource = versioned_spec.resources[resource.capitalize()]
            except KeyError:
                raise AttributeError(
                    'Resource {0} not found on API revision {1}'.format(
                        resource, resource_version
                    )
                )
            base_spec.resources[resource.capitalize()] = spec_resource
    return base_spec


def read_spec(path, http_client=None):
    """
    Reads in a swagger spec file used to initialize a SwaggerClient
    :param path: String path to local swagger spec file.
    :param http_client: :class:`bravado.requests_client.RequestsClient`
    :return: :class:`bravado_core.spec.Spec`
    """
    with open(path, 'r') as f:
        spec_dict = json.loads(f.read())

    return SwaggerClient.from_spec(
        spec_dict, http_client=http_client, config=SPEC_CONFIG
    )


def esi_client_factory(
    token=None, datasource=None, spec_file=None, version=None, **kwargs
):
    """
    Generates an ESI client.
    :param token: :class:`esi.Token` used to access authenticated endpoints.
    :param datasource: Name of the ESI datasource to access.
    :param spec_file: Absolute path to a swagger spec file to load.
    :param version: Base ESI API version. Accepted values are 'legacy', 'latest', 
    'dev', or 'vX' where X is a number.
    :param kwargs: Explicit resource versions to build, in the form Character='v4'. 
    Same values accepted as version.
    :return: :class:`bravado.client.SwaggerClient`

    If a spec_file is specified, specific versioning is not available. 
    Meaning the version and resource version kwargs
    are ignored in favour of the versions available in the spec_file.
    """    
    if app_settings.ESI_INFO_LOGGING_ENABLED:
        logger.info('Generating an ESI client...')
    client = requests_client.RequestsClient()
    my_http_adapter = HTTPAdapter(
        pool_maxsize=app_settings.ESI_CONNECTION_POOL_MAXSIZE, 
        max_retries=app_settings.ESI_CONNECTION_ERROR_MAX_RETRIES
    )
    client.session.mount('https://', my_http_adapter)
    
    if token or datasource:
        client.authenticator = TokenAuthenticator(token=token, datasource=datasource)

    api_version = version or app_settings.ESI_API_VERSION

    if spec_file:
        return read_spec(spec_file, http_client=client)
    else:
        spec = build_spec(api_version, http_client=client, **kwargs)
        return SwaggerClient(spec)


def minimize_spec(spec_dict, operations=None, resources=None):
    """
    Trims down a source spec dict to only the operations or resources indicated.
    :param spec_dict: The source spec dict to minimize.
    :type spec_dict: dict
    :param operations: A list of operation IDs to retain.
    :type operations: list of str
    :param resources: A list of resource names to retain.
    :type resources: list of str
    :return: Minimized swagger spec dict
    :rtype: dict
    """
    operations = operations or []
    resources = resources or []

    # keep the ugly overhead for now but only add paths we need
    minimized = {key: value for key, value in spec_dict.items() if key != 'paths'}
    minimized['paths'] = {}

    for path_name, path in spec_dict['paths'].items():
        for method, data in path.items():
            if (
                data['operationId'] in operations 
                or any(tag in resources for tag in data['tags'])
            ):
                if path_name not in minimized['paths']:
                    minimized['paths'][path_name] = {}
                minimized['paths'][path_name][method] = data

    return minimized


class EsiClientProvider:
    """Class for providing a single ESI client instance for the whole app"""

    _client = None
    
    def __init__(self, datasource=None, spec_file=None, version=None, **kwargs):
        """        
        :param datasource: Name of the ESI datasource to access.
        :param spec_file: Absolute path to a swagger spec file to load.
        :param version: Base ESI API version. 
        Accepted values are 'legacy', 'latest', 'dev', or 'vX' where X is a number.
        :param kwargs: Explicit resource versions to build, 
        in the form Character='v4'. Same values accepted as version.        

        If a spec_file is specified, specific versioning is not available. 
        Meaning the version and resource version kwargs
        are ignored in favour of the versions available in the spec_file.
        """
        self._datasource = datasource
        self._spec_file = spec_file
        self._version = version
        self._kwargs = kwargs

    @property
    def client(self):
        if self._client is None:
            self._client = esi_client_factory(
                datasource=self._datasource,                
                spec_file=self._spec_file,
                version=self._version,
                **self._kwargs,
            )
        return self._client

    def __str__(self):
        return 'EsiClientProvider'



# -*- coding: utf-8 -*-
# Generated by Django 1.10.5 on 2017-03-30 01:43
from __future__ import unicode_literals

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('esi', '0003_hide_tokens_from_admin_site'),
    ]

    operations = [
        migrations.AlterField(
            model_name='token',
            name='access_token',
            field=models.CharField(editable=False, help_text='The access token granted by SSO.', max_length=254),
        ),
    ]



import logging
from functools import wraps

from .models import Token, CallbackRedirect


logger = logging.getLogger(__name__)


def _check_callback(request):
    # ensure session installed in database
    if not request.session.exists(request.session.session_key):
        logger.debug("Creating new session for %s", request.user)
        request.session.create()

    # clean up callback redirect, pass token if new requested
    try:
        model = CallbackRedirect.objects.get(session_key=request.session.session_key)
        token = Token.objects.get(pk=model.token.pk)
        model.delete()
        logger.debug(
            "Retrieved new token from callback for %s session %s", 
            request.user, 
            request.session.session_key[:5])
        return token
    
    except (CallbackRedirect.DoesNotExist, Token.DoesNotExist, AttributeError):
        logger.debug(
            "No callback for %s session %s",
            request.user, 
            request.session.session_key[:5],
            exc_info=True
        )
        return None


def tokens_required(scopes='', new=False):
    """
    Decorator for views to request an ESI Token.
    Accepts required scopes as a space-delimited string
    or list of strings of scope names.
    Can require a new token to be retrieved by SSO.
    Returns a QueryDict of Tokens.
    """

    def decorator(view_func):
        @wraps(view_func)
        def _wrapped_view(request, *args, **kwargs):

            # if we're coming back from SSO for a new token, return it
            token = _check_callback(request)
            if token and new:
                tokens = Token.objects.filter(pk=token.pk)
                logger.debug("Returning new token.")
                return view_func(request, tokens, *args, **kwargs)

            if not new:
                # ensure user logged in to check existing tokens
                if not request.user.is_authenticated:
                    logger.debug(
                        "Session %s is not logged in. Redirecting to login.", 
                        request.session.session_key[:5]
                    )
                    from django.contrib.auth.views import redirect_to_login
                    return redirect_to_login(request.get_full_path())

                # collect tokens in db, check if still valid, return if any
                tokens = (
                    Token.objects
                    .filter(user__pk=request.user.pk)
                    .require_scopes(scopes)
                    .require_valid()
                )
                if tokens.exists():
                    logger.debug(
                        "Retrieved %s tokens for %s session %s",
                        tokens.count(), 
                        request.user,
                        request.session.session_key[:5]
                    )
                    return view_func(request, tokens, *args, **kwargs)

            # trigger creation of new token via sso
            logger.debug(
                "No tokens identified for %s session %s. Redirecting to SSO.",
                request.user,
                request.session.session_key[:5]
            )
            from esi.views import sso_redirect
            return sso_redirect(request, scopes=scopes)

        return _wrapped_view

    return decorator


def token_required(scopes='', new=False):
    """
    Decorator for views which supplies a single, 
    user-selected token for the view to process.
    Same parameters as tokens_required.
    """

    def decorator(view_func):
        @wraps(view_func)
        def _wrapped_view(request, *args, **kwargs):

            # if we're coming back from SSO for a new token, return it
            token = _check_callback(request)
            if token and new:
                logger.debug(
                    "Got new token from %s session %s. Returning to view.", 
                    request.user, 
                    request.session.session_key[:5]
                )
                return view_func(request, token, *args, **kwargs)

            # if we're selecting a token, return it
            if request.method == 'POST':
                if request.POST.get("_add", False):
                    logger.debug(
                        "%s has selected to add new token. Redirecting to SSO.",
                        request.user
                    )
                    # user has selected to add a new token
                    from esi.views import sso_redirect
                    return sso_redirect(request, scopes=scopes)

                token_pk = request.POST.get('_token', None)
                if token_pk:
                    logger.debug(
                        "%s has selected token %s", request.user, token_pk
                    )
                    try:
                        token = Token.objects.get(pk=token_pk)
                        # ensure token belongs to this user and has required scopes
                        if (
                            (
                                (token.user and token.user == request.user) 
                                or not token.user
                            ) 
                            and Token.objects
                            .filter(pk=token_pk)
                            .require_scopes(scopes)
                            .require_valid()
                            .exists()
                        ):
                            logger.debug(
                                "Selected token fulfills requirements of view. "
                                "Returning."
                            )
                            return view_func(request, token, *args, **kwargs)

                    except Token.DoesNotExist:
                        logger.debug("Token %s not found.", token_pk)                       

            if not new:
                # present the user with token choices
                tokens = (
                    Token.objects
                    .filter(user__pk=request.user.pk)
                    .require_scopes(scopes)
                    .require_valid()
                )
                if tokens.exists():
                    logger.debug(
                        "Returning list of available tokens for %s.", request.user
                    )
                    from esi.views import select_token
                    return select_token(request, scopes=scopes, new=new)
                else:
                    logger.debug(
                        "No tokens found for %s session %s with scopes %s",
                        request.user, 
                        request.session.session_key[:5],
                        scopes
                    )

            # prompt the user to add a new token
            logger.debug(
                "Redirecting %s session %s to SSO.", 
                request.user, 
                request.session.session_key[:5]
            )
            from esi.views import sso_redirect
            return sso_redirect(request, scopes=scopes)

        return _wrapped_view

    return decorator


def single_use_token(scopes='', new=False):
    """
    Decorator for views which supplies a single use token granted via sso login 
    regardless of login state.
    Same parameters as tokens_required.
    """

    def decorator(view_func):
        @wraps(view_func)
        def _wrapped_view(request, *args, **kwargs):

            # if we're coming back from SSO for a new token, return it
            token = _check_callback(request)
            if token:
                logger.debug(
                    "Got new token from session %s. Returning to view.",
                    request.session.session_key[:5]
                )
                return view_func(request, token, *args, **kwargs)

            # prompt the user to login for a new token
            logger.debug(
                "Redirecting session %s to SSO.", request.session.session_key[:5]
            )
            from esi.views import sso_redirect
            return sso_redirect(request, scopes=scopes)

        return _wrapped_view

    return decorator



from django.contrib import admin
from django.contrib.auth import get_user_model

from .models import Token, Scope, CallbackRedirect

admin.site.register(CallbackRedirect)


@admin.register(Scope)
class ScopeAdmin(admin.ModelAdmin):
    list_display = ('name', 'help_text')


@admin.register(Token)
class TokenAdmin(admin.ModelAdmin):
    def get_scopes(self, obj):
        return ", ".join([x.name for x in obj.scopes.all()])

    get_scopes.short_description = 'Scopes'

    User = get_user_model()
    list_display = ('user', 'character_name', 'get_scopes')
    search_fields = ['user__%s' % User.USERNAME_FIELD, 'character_name', 'scopes__name']



# -*- coding: utf-8 -*-
# Generated by Django 1.10.3 on 2016-11-10 15:12
from __future__ import unicode_literals

from django.db import migrations

SCOPES = {
    'esi-planets.manage_planets.v1':
        "Allows reading a list of a character's planetary colonies, and the details of those colonies.",
    'esi-ui.open_window.v1': "Allows open window in game client remotely.",
    'esi-assets.read_assets.v1': "Allows reading a list of assets that the character owns.",
    'esi-calendar.read_calendar_events.v1': "Allows reading a character's calendar, including corporate events.",
    'esi-bookmarks.read_character_bookmarks.v1': "Allows reading of a character's bookmarks and bookmark folders.",
    'esi-wallet.read_character_wallet.v1': "Allows reading of a character's wallet, journal and transaction history.",
    'esi-clones.read_clones.v1': "Allows reading the locations of a character's jump clones and their implants.",
    'esi-characters.read_contacts.v1':
        "Allows reading of a character's contacts list, and calculation of CSPA charges.",
    'esi-corporations.read_corporation_membership.v1':
        "Allows reading a list of the ID's and roles of a character's fellow corporation members.",
    'esi-fleets.read_fleet.v1': "Allows reading information about fleets.",
    'esi-killmails.read_killmails.v1': "Allows reading of a character's kills and losses.",
    'esi-location.read_location.v1': "Allows reading of a character's active ship location.",
    'esi-location.read_ship_type.v1': "Allows reading of a character's active ship class.",
    'esi-skills.read_skillqueue.v1': "Allows reading of a character's currently training skill queue.",
    'esi-skills.read_skills.v1': "Allows reading of a character's currently known skills.",
    'esi-universe.read_structures.v1':
        "Allows querying the location and type of structures that the character has docking access at.",
    'esi-calendar.respond_calendar_events.v1': "Allows updating of a character's calendar event responses.",
    'esi-search.search_structures.v1':
        "Allows searching over all structures that a character can see in the structure browser.",
    'esi-fleets.write_fleet.v1': "Allows manipulating fleets.",
    'esi-ui.write_waypoint.v1': "Allows manipulating waypoints in game client remotely."
}


def generate_scopes(apps, schema_editor):
    Scope = apps.get_model('esi', 'Scope')
    for s in SCOPES:
        Scope.objects.update_or_create(name=s, defaults={'help_text': SCOPES[s]})


def delete_scopes(apps, schema_editor):
    Scope = apps.get_model('esi', 'Scope')
    for s in SCOPES:
        try:
            Scope.objects.get(name=s).delete()
        except Scope.DoesNotExist:
            pass


class Migration(migrations.Migration):
    dependencies = [
        ('esi', '0001_initial'),
    ]

    operations = [
        migrations.RunPython(generate_scopes, delete_scopes)
    ]



# Generated by Django 2.0.4 on 2018-04-30 00:06

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('esi', '0004_remove_unique_access_token'),
    ]

    operations = [
        migrations.AlterField(
            model_name='token',
            name='access_token',
            field=models.TextField(editable=False, help_text='The access token granted by SSO.'),
        ),
        migrations.AlterField(
            model_name='token',
            name='refresh_token',
            field=models.TextField(blank=True, default='', editable=False, help_text='A re-usable token to generate new access tokens upon expiry.'),
        ),
    ]



# Generated by Django 2.2.9 on 2020-04-14 13:17

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('esi', '0005_remove_token_length_limit'),
    ]

    operations = [
        migrations.AlterField(
            model_name='callbackredirect',
            name='url',
            field=models.TextField(default='/', help_text='The internal URL to redirect this callback towards.'),
        ),
    ]



import os
from setuptools import find_packages, setup
from esi import __version__

this_directory = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

setup(
    name='django-esi',
    version=__version__,
    python_requires='~=3.6',
    install_requires=[
        'requests>=2.9.1,<3.0',
        'requests_oauthlib>=0.8.0,<2.0',
        'django>=2.0',
        'bravado>=10.6.0,<11.0',
        'celery>=4.0.2',
    ],
    packages=find_packages(),
    include_package_data=True,
    license='GNU General Public License v3 (GPLv3)',
    description='Django app for accessing the EVE Swagger Interface (ESI).',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://gitlab.com/allianceauth/django-esi',
    author='Alliance Auth',
    author_email='adarnof@gmail.com',
    classifiers=[
        'Environment :: Web Environment',
        'Framework :: Django',        
        'Framework :: Django :: 2.2',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',        
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
    ],
)



from django.core.checks import Error, Warning, register, Tags
from django.conf import settings


@register(Tags.security)
def check_sso_application_settings(*args, **kwargs):
    errors = []
    try:
        assert (
            settings.ESI_SSO_CLIENT_ID 
            and settings.ESI_SSO_CLIENT_SECRET 
            and settings.ESI_SSO_CALLBACK_URL
        )
    except (AssertionError, AttributeError):
        if settings.DEBUG:
            errors.append(
                Warning(
                    'ESI SSO application settings are not configured.', 
                    hint='SSO features will not work.',
                    id='esi.W001'
                )
            )
        else:
            errors.append(
                Error(
                    'ESI SSO application settings cannot be blank.',
                    hint=(
                        'Register an application at https://developers.eveonline.com '
                        'and add ESI_SSO_CLIENT_ID, ESI_SSO_CLIENT_SECRET, and '
                        'ESI_SSO_CALLBACK_URL to your project settings.'
                    ),
                    id='esi.E001'
                )
            )
    return errors



class DjangoEsiException(Exception):
    pass


class TokenError(DjangoEsiException):
    pass


class TokenInvalidError(TokenError):
    pass


class TokenExpiredError(TokenError):
    pass


class NotRefreshableTokenError(TokenError):
    pass


class IncompleteResponseError(DjangoEsiException):
    pass
